{ File   : R000Ext.dpr
  Author : Deniz Oezmen
  Created: 2005-10-09
  Changed: 2005-10-10

  Runaway 000 extractor.
}
program R000Ext;

{$APPTYPE CONSOLE}

uses
  SysUtils, StrUtils;

type
  TR000Header = packed record
                  Offsets,
                  Sizes  :array[0..5999] of Longword
                end;

  TBuffer     = array of Byte;

var                                                    
  R000Header: TR000Header;
  R000File  : file;
  IsWAV     : Boolean;
  i         : Longword;

function IntToStrL(i, l: Integer): string;
begin
  Result := IntToStr(i);

  while Length(Result) < l do
    Result := '0' + Result
end;

function Prefix(FileName: string): string;
var
  i: Longint;
begin
  FileName := ExpandFileName(FileName);

  i := Length(FileName) + 1;
  repeat
    Dec(i)
  until (i = 0) or (PosEx('.', FileName, i) > 0);

  if (i = 0) or (PosEx('\', FileName, i + 1) > 0) then
    Result := FileName
  else
    Result := Copy(FileName, 1, i - 1)
end;

function OpenR000File(var R000File: file; R000Name: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(R000File, R000Name);

  {$I-}
  Reset(R000File, 1);
  {$I+}
  OpenR000File := (IOResult = 0)
end;

function ReadHeader: Boolean;
var
  BytesRead: Longword;
begin
  BlockRead(R000File, R000Header, SizeOf(TR000Header), BytesRead);
  ReadHeader := (BytesRead = SizeOf(TR000Header))
end;

procedure ExtractFiles(R000Name: string);
var
  i,
  Ctr      : Longword;
  AreaFound: Boolean;
  OutFile  : file;
  Buffer   : TBuffer;
begin
  // remove non-fitting entries (odd stuff ...)
  AreaFound := False;
  for i := 5999 downto 0 do
  begin
    if not AreaFound then
    begin
      if R000Header.Sizes[i] = 0 then
        R000Header.Offsets[i] := 0;
      AreaFound := (R000Header.Offsets[i] = 48000)
    end
    else
    begin
      R000Header.Sizes[i] := 0;
      R000Header.Offsets[i] := 0
    end
  end;

  Ctr := 0;
  FileMode := fmOpenReadWrite;

  for i := 0 to 5999 do
  begin
    if R000Header.Offsets[i] = 0 then
      Continue;

    WriteLn('  #', IntToStrL(Ctr, 4));

    Assign(OutFile, Prefix(R000Name) + '#' + IntToStrL(Ctr, 4) + '.dat');
    Rewrite(OutFile, 1);

    SetLength(Buffer, R000Header.Sizes[i]);
    BlockRead(R000File, Buffer[0], R000Header.Sizes[i]);
    BlockWrite(OutFile, Buffer[0], R000Header.Sizes[i]);

    Close(OutFile);

    Inc(Ctr)
  end;
end;

procedure ParseFiles(FilePara: string);
var
  DirPos,
  i        : Byte;
  FileName,
  Dir,
  R000Name  : string;
  SearchRec: TSearchRec;
  PFilePara: PChar;
begin
  GetMem(PFilePara, Length(FilePara) + 1);
  StrPCopy(PFilePara, FilePara);

  // get the directory information from FilePara
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  FileMode := fmOpenRead;

  // search for all files matching FilePara
  if FindFirst(PFilePara, faAnyFile - faDirectory - faVolumeID,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      R000Name := Dir + FileName;

      // if this file is accessible, analyze it
      if OpenR000File(R000File, R000Name) then
        if ReadHeader then
        begin
          Writeln('Extracting from ', R000Name, ' ...');
          ExtractFiles(R000Name);
          Close(R000File)
        end
        else
          Writeln(R000Name, ' is not a valid 000 file.')
      else
        Writeln(R000Name, ' not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    Writeln(FilePara, ' not found')
end;

begin
  if ParamCount >= 1 then
  begin
    IsWAV := False;
    for i := 2 to ParamCount do
      IsWAV := IsWAV or (UpperCase(ParamStr(i)) = '/WAV');
    ParseFiles(ParamStr(1))
  end
  else
    Writeln('Incorrect number of parameters.')
end.

