{ File   : RSExt.dpr
  Author : Deniz Oezmen
  Created: 2005-10-10
  Changed: 2005-10-10

  Runaway S?? extractor.
}
program RSExt;

{$APPTYPE CONSOLE}

uses
  SysUtils, StrUtils;

type
  TRSHeader = packed record
                  Offsets: array[0..999] of Longword
                end;

  TBuffer   = array of Byte;

var
  RSHeader: TRSHeader;
  RSFile  : file;
  IsWAV   : Boolean;
  i       : Longword;

function IntToStrL(i, l: Integer): string;
begin
  Result := IntToStr(i);

  while Length(Result) < l do
    Result := '0' + Result
end;

function Prefix(FileName: string): string;
var
  i: Longint;
begin
  FileName := ExpandFileName(FileName);

  i := Length(FileName) + 1;
  repeat
    Dec(i)
  until (i = 0) or (PosEx('.', FileName, i) > 0);

  if (i = 0) or (PosEx('\', FileName, i + 1) > 0) then
    Result := FileName
  else
    Result := Copy(FileName, 1, i - 1)
end;

function OpenRSFile(var RSFile: file; RSName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(RSFile, RSName);

  {$I-}
  Reset(RSFile, 1);
  {$I+}
  OpenRSFile := (IOResult = 0)
end;

function ReadHeader: Boolean;
var
  BytesRead: Longword;
begin
  BlockRead(RSFile, RSHeader, SizeOf(TRSHeader), BytesRead);
  ReadHeader := (BytesRead = SizeOf(TRSHeader))
end;

procedure ExtractFiles(RSName: string);
var
  i,
  Ctr,
  Ofs,
  Size     : Longword;
  OutFile  : file;
  Buffer   : TBuffer;
begin
  Ctr := 0;
  FileMode := fmOpenReadWrite;

  i := 1;
  while i <= 999 do
  begin
    Ofs := RSHeader.Offsets[i];

    if Ofs = 0 then
      Break;

    while (RSHeader.Offsets[i] <= Ofs) and (i < 999) do
      Inc(i);

    if RSHeader.Offsets[i] <= Ofs then
      Break;

    Size := RSHeader.Offsets[i] - Ofs;

    WriteLn('  ', '#' + IntToStrL(Ctr, 4));

    Assign(OutFile, Prefix(RSName) + '#' + IntToStrL(Ctr, 4) + '.dat');
    Rewrite(OutFile, 1);

    SetLength(Buffer, Size);
    Seek(RSFile, Ofs);
    BlockRead(RSFile, Buffer[0], Size);
    BlockWrite(OutFile, Buffer[0], Size);

    Inc(Ctr);

    Close(OutFile)
  end;
end;

procedure ParseFiles(FilePara: string);
var
  DirPos,
  i        : Byte;
  FileName,
  Dir,
  RSName  : string;
  SearchRec: TSearchRec;
  PFilePara: PChar;
begin
  GetMem(PFilePara, Length(FilePara) + 1);
  StrPCopy(PFilePara, FilePara);

  // get the directory information from FilePara
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  FileMode := fmOpenRead;

  // search for all files matching FilePara
  if FindFirst(PFilePara, faAnyFile - faDirectory - faVolumeID,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      RSName := Dir + FileName;

      // if this file is accessible, analyze it
      if OpenRSFile(RSFile, RSName) then
        if ReadHeader then
        begin
          Writeln('Extracting from ', RSName, ' ...');
          ExtractFiles(RSName);
          Close(RSFile)
        end
        else
          Writeln(RSName, ' is not a valid S?? file.')
      else
        Writeln(RSName, ' not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    Writeln(FilePara, ' not found')
end;

begin
  if ParamCount >= 1 then
  begin
    IsWAV := False;
    for i := 2 to ParamCount do
      IsWAV := IsWAV or (UpperCase(ParamStr(i)) = '/WAV');
    ParseFiles(ParamStr(1))
  end
  else
    Writeln('Incorrect number of parameters.')
end.

